library(modreg)
iqr <- function(x) as.numeric(diff(quantile(x,c(1,3)/4)))

convert <- function(file,
                    path=paste(.path.package(package="affy"),"/exec/",sep=""),
                    read=T,chip.names=NULL,verbose=T){
  if(missing(file)) stop("Argument file missing, with no default\n")
  if(verbose) cat("Converting cel files to table... this may take a few minutes\n")
  system(paste(path,"cel2tab.pl ",file,sep=""))
  if(read){
    
    if(verbose) cat("Reading in data.\n")
    return(read.affy(file,system("ls *.info",TRUE),chip.names))
  }
  else
    return(NULL)
}
   
read.affy <- function(file,chip,chip.names=NULL){
  Names <- scan(file,nlines=1,what="c")
  nchips <- length(Names) - 5 ##first five are info not intensities
  if(is.null(chip.names)) chip.names <- Names[-c(1:5)]
  else{
    if(length(chip.names)!=nchips){
      warning("Not the same number of chips than chip names. Assigning names from file.\n") 
      chip.names <- Names[-c(1:5)]
    }
  }
  Data <- matrix(scan(file,skip=1),ncol=length(Names),byrow=T)
  ncol <- max(Data[,4])+1
  nrow <- max(Data[,5])+1
  n <- dim(Data)[1]
  pmindex <- (1:n)[Data[,3] == 1]
  mmindex <- (1:n)[Data[,3] == 0]
  if(any((pmindex+ncol)-mmindex))
    warning("MMs are not underneath PMs\n")
  pm <-matrix(as.matrix(Data[pmindex,-c(1:5)]),ncol=nchips)##contains all pms
  mm <-matrix(as.matrix(Data[mmindex,-c(1:5)]),ncol=nchips)##contains all mms
  dimnames(pm)<- list(NULL,chip.names)
  dimnames(mm)<- list(NULL,chip.names)
  probe.names <- Data[pmindex,1]
  x <- Data[pmindex,4]
  y <- Data[pmindex,5]
  probe.numbers <- Data[pmindex,2]
  if(any(pm<=0)){
    warning("Non positive pms found. Substituting with minimum.\n")
     pm[pm<=0] <- min(pm[pm>0])
   }
  if(any(mm<=0)){
    warning("Non positive mms found. Substituting with minimum.\n")
    mm[mm<=0] <- min(mm[mm>0])
  }
  probe.names <- scan(chip,what="c")
  return(list(pm=pm,mm=mm,id=Data[pmindex,1],
              numbers=probe.numbers,names=probe.names,n=length(probe.names),
              x=x,y=y,ncol=ncol,nrow=nrow,nchips=nchips,
              nprob=length(pmindex),chip.names=chip.names))
}


    


#
# parts of code copied from code written by Cheng Li, Fred A. Wright, Ben Bolstad
#
### Implementation of reduced model of Li C, Wong WH PNAS 2001 98:31-36
fit.li.wong <- function(data.matrix, remove.outliers=T,
                        normal.array.quantile=0.5,
                        normal.resid.quantile=0.9,
                        large.threshold=3,
                        large.variation=0.8,
                        outlier.fraction=0.14,
                        delta = 1e-06,maxit=50,outer.maxit=50,verbose=F){
  if(missing(data.matrix)) stop("Argument data.matrix missing, with no default")
  II <- dim(data.matrix)[1] ##II instrad of I cause I is a fuction in R
  J <- dim(data.matrix)[2]
  cI <- II ##current I
  cJ <- J ##current J
  theta.outliers.old <- rep(F,II) ##ith entry will be true if theta_i is an outlier
  phi.outliers.old <- rep(F,J) ##jth entry will be true if phi_j is an outlier
  single.outliers.old <- matrix(F,II,J) ##ij entry will be true if y_is an outlier
  theta.outliers <- theta.outliers.old ##need this to now if change ocurred in outliers
  phi.outliers <- phi.outliers.old ##need this to know if chages occured in outlies
  single.outliers <- single.outliers.old
  flag1 <- NA ##these will be false if convergence not reacher,
  flag2 <- NA ## this will be false if outliers respectively cuase iter to stop
  if(remove.outliers){
    flag1 <- T;flag2<-T
    original.data.matrix <- data.matrix ##so we can get it back after outlier removal
    change.theta <- 1 #start with 1 
    change.phi <- 1
    change.single <- 1
    outer.iter <- 0 
    while(flag1 & flag2 & change.theta+change.phi+change.single >0 & outer.iter < outer.maxit) {
      outer.iter <- outer.iter + 1
      
      if((outer.iter%%3==0 & change.theta>0)  |
         (outer.iter%%3==1 & change.phi>0)){ #something has to change
        ##starting values
        phi <- apply(data.matrix, 2, mean)
        c <- sqrt(cJ/sum(phi[!phi.outliers]^2))
        phi <- c * phi
        theta <- (data.matrix[,!phi.outliers] %*% phi[!phi.outliers])/cJ
        iter <- 0
        change <- 1 #start with one
        theta.old <- rep(0, II)
        while(change > delta & iter < maxit) {
          iter <- iter + 1
          phi <- t(data.matrix[!theta.outliers,]) %*% theta[!theta.outliers] ##ignore the outliers
          c <- sqrt(cJ/sum(phi[!phi.outliers]^2))
          phi <- c * phi
          theta <- (data.matrix[,!phi.outliers] %*% phi[!phi.outliers])/cJ
          change <- max(abs(theta[!theta.outliers] - theta.old[!theta.outliers]))
          if(verbose) cat(paste("Outlier iteration:",outer.iter,"estimation iteration:",iter,"chage=",change,"\n"))
          theta.old <- theta
        }
        if(iter>=maxit){ ##convergence not reached. might as well get out
          warning(paste("No convergence in inner loop after",iter,"in outerler tieration",outer.iter,"\n"))
          flag1 <- F
        }
        if(mean(phi[!phi.outliers]<0)>.5){ ##for identifiability.. theta*phi = (-theta)*(-phi), i require that most phis are positive
          theta <- -theta
          phi <- -phi
        }
        theta <- as.vector(theta)
        phi <- as.vector(phi)
                
        data.matrixhat <- outer(theta, phi)
        resid <- data.matrix-data.matrixhat
      }
      ##DEALING WITH OUTLIERS
      ##we alternate removal of outliers
      ##if even iteration take out thetas that are outliers (as defined by Li and Wong).
      if(outer.iter%%3==1){ ## we start with single outliers
        single.outliers <- resid > large.threshold*quantile(abs(resid),normal.resid.quantile)
        single.outliers[apply(single.outliers,1,sum) > outlier.fraction*cJ,]<-rep(F,J)
        ##probably chip oulier, defer calling outlier
        single.outliers[,apply(single.outliers,2,sum) > outlier.fraction*cI]<-rep(F,II)
        ##probably probe outlier, defer calling outlier
        data.matrix[single.outliers] <- data.matrixhat[single.outliers]
        data.matrix[!single.outliers] <- original.data.matrix[!single.outliers]
        change.single <- sum(abs(single.outliers.old-single.outliers)) #sum will be total of changes
        single.outliers.old <- single.outliers
      }
      else{
        sigma.theta <- sqrt(apply(resid[,!phi.outliers]^2, 1, sum)/(cJ - 1))
        sigma.phi <- sqrt(apply(resid[!theta.outliers,]^2, 2, sum)/(cI - 1))
        ###THETA OUTLIERS
        if(outer.iter%%3==2){
          theta.outliers <- sigma.theta > large.threshold*quantile(sigma.theta,normal.array.quantile) | theta^2/sum(theta^2) > large.variation
          cI <- sum(!theta.outliers)
          if(cI<3) {
            warning("No convergence achieved, too many outliers")
            flag2 <- F
          }
          ##single outliers in outlier chips are not longer single outliers
          single.outliers[theta.outliers,] <- rep(F,J)
          data.matrix[single.outliers] <- data.matrixhat[single.outliers]
          data.matrix[!single.outliers]<-original.data.matrix[!single.outliers]
          change.theta <- sum(abs(theta.outliers.old-theta.outliers)) #sum will be total of changes
          change.single <- sum(abs(single.outliers.old-single.outliers)) #sum will be total of changes
          theta.outliers.old <- theta.outliers
        }
        ##PHI OUTLIERS
        else{
          phi.outliers <- sigma.phi > large.threshold*quantile(sigma.phi,normal.array.quantile) | phi^2/sum(phi^2) > large.variation | phi <0
          cJ <- sum(!phi.outliers)
          if(cJ<3) {
            warning("No convergence achieved, too many outliers")
            flag2 <- F
          }
          single.outliers[,phi.outliers] <- rep(F,II)
          data.matrix[single.outliers] <- data.matrixhat[single.outliers]
          data.matrix[!single.outliers]<-original.data.matrix[!single.outliers]
          change.phi <- sum(abs(phi.outliers.old-phi.outliers))
          change.single <- sum(abs(single.outliers.old-single.outliers)) #sum will be total of changes
          phi.outliers.old <- phi.outliers
        }
      }
      if(verbose){
        cat("chips used=",cI,", probes used=",cJ,", single outler=",sum(single.outliers),"\n")
        cat("Number of changes: single=",change.single,", theta=",change.theta,", phi=",change.phi,"\n",sep="")
      }
    }
    if(outer.iter>=outer.maxit){
      warning("No convergence achieved in outlier loop\n")
      flag2 <- F
    }
    all.outliers <- outer(theta.outliers,phi.outliers,FUN="|") | single.outliers
    sigma <- sqrt(sum(resid[!all.outliers]^2)/sum(!all.outliers))
    ##in case we leave iteration and these havent been defined
    sigma.theta <- sqrt(apply(resid[,!phi.outliers]^2, 1, sum)/(cJ - 1))
    sigma.phi <- sqrt(apply(resid[!theta.outliers,]^2, 2, sum)/(cI - 1))
  }
  ###code for NO OUTLIER REMOVAL
  else{
    flag1 <- T
    phi <- apply(data.matrix, 2, mean)
    c <- sqrt(J/sum(phi^2))
    phi <- c * phi
    theta <- (data.matrix %*% phi)/J

    iter <- 0
    change <- 1
    theta.old <- rep(0, II)
    while(change > delta & iter < maxit) {
      iter <- iter + 1
      phi <- t(data.matrix) %*% theta
      c <- sqrt(J/sum(phi^2))
      phi <- c * phi
      theta <- (data.matrix %*% phi)/J
      change <- max(abs(theta - theta.old))
      if(verbose) cat(paste("Iteration:",iter,"chage=",change,"\n"))
      theta.old <- theta
    }
    if(iter>=maxit){
      warning(paste("No convergence after",iter,"iterations.\n"))
      flag1 <- F
    }
    if(mean(phi[!phi.outliers]<0)>.5){
      ##for identifiability.. theta*phi = (-theta)*(-phi), i require that most phis are positive
      theta <- -theta
      phi <- -phi
    }
    theta <- as.vector(theta)
    phi <- as.vector(phi)
    data.matrixhat <- outer(theta, phi)
    sigma.theta <- sqrt(apply((data.matrix - data.matrixhat)^2, 1, sum)/(J - 1))
    sigma.phi <- sqrt(apply((data.matrix - data.matrixhat)^2, 2, sum)/(II - 1))
    sigma <- sqrt(sum((data.matrix - data.matrixhat)^2)/(II * J))
  }
  return(list(theta = theta, phi = phi, sigma.eps = sigma, sigma.theta = sigma.theta, sigma.phi=sigma.phi,theta.outliers=theta.outliers,phi.outliers=phi.outliers,single.outliers=single.outliers,convergence1=flag1,convergence2=flag2,iter = iter, delta = change)) 
}


li.wong <- function(x,subset=NULL,ignore.mm=F,remove.outliers=T,
                    normal.array.quantile=0.5,
                    normal.resid.quantile=0.9,
                    large.threshold=3,
                    large.variation=0.8,
                    outlier.fraction=0.14,
                    delta = 1e-06,maxit=50,outer.maxit=50,verbose=F){
  if(missing(x)) stop("Argument x missing, with no default\n")
  if(ignore.mm)
    aux <- x$pm
  else
    aux <- x$pm-x$mm
  unique.probe.names <- x$names
  if(!is.null(subset)) unique.probe.names <- unique.probe.names[subset]
  n <- x$n
  theta <- matrix(0,n,x$nchips)
  sigma.theta <- matrix(0,n,x$nchips)
  outlier <- matrix(T,n,x$nchips)
  sigma.eps <- rep(0,n)
  convergence1 <- rep(T,n)
  convergence2 <- rep(T,n)
  for(i in 1:n){ 
    index <- x$id==i
    res <- fit.li.wong(t(aux[index,]),remove.outliers,normal.array.quantile,normal.resid.quantile,large.threshold,large.variation,outlier.fraction,delta,maxit,outer.maxit,verbose)
    theta[i,] <- res$theta
    sigma.theta[i,] <- res$sigma.theta
    sigma.eps[i] <- res$sigma.eps
    outlier[i,] <- res$theta.outliers
    convergence1[i] <- res$convergence1
    convergence2[i] <- res$convergence2
  }
  return(list(names=unique.probe.names,theta=theta,sigma.theta=sigma.theta,sigma.eps=sigma.eps,outlier=outlier,convergence1=convergence1,convergence2=convergence2))
}

avdiff <- function(x,verbose=F){
  if(missing(x)) stop("Argument x missing, with no default\n")
  cat("Computing average difference for",dim(x$pm)[2],"columns")
  avdiff <- apply(as.matrix(x$pm-x$mm),2,function(y){
    cat(".")
    tapply(y,x$id,function(z){
      o <- order(z)
      zz <- z[-c(o[1],o[length(z)])] #take out biggest and smallest
      mean(z[abs(z-mean(zz))<3*sd(zz)])
    })
  })
  dimnames(avdiff) <- list(x$names,x$chip.names)
  return(avdiff)
}

affy.subset <- function(x,probes=NULL,chips=NULL,genes=NULL){
  if(is.null(chips) & is.null(probes) & is.null(genes)) return(x)
  else{
    if(is.null(chips)) chips <- 1:x$nchips
    if(is.null(genes)){
      flag <- F ##to avoid intersect
      probe.names <- x$names
      genes <- 1:x$nprob
    }
    else{
      aux <- c()
      for(i in genes){
        aux <- c(aux,(1:x$nprob)[x$id==i])
      }        
      genes <- aux
      flag <- T
    }
    if(is.null(probes)) probes <- genes
    else{
      if(flag){
        probes <- intersect(genes,1:x$nprob)
      }
    }        
    return(pm=x$pm[probes,chips],mm=x$mm[probes,chips],
           numbers=x$numbers[probes],
           names=x$names[unique(x$id[probes])],
           n=length(x$names[unique(x$id[probes])]),
           x=x$x[probes],y=x$y[probes],ncol=x$ncol,
           nrow=x$nrow,
           nchips=length(chips),nprob=length(probes),
           chip.names=x$chip.names[chips])
  }
}

affy.mva.pairs <- function(x,chips=c(1:x$nchips),subset=NULL,
                           labels=x$chip.names,
                           span=2/3,family.loess="symmetric",
                           digits=3,line.col=2,subset.size=2000,
                           main="MVA plot",...){
  alldata  <- rbind(x$pm[,chips],x$mm[,chips])
  if(is.null(subset))
    subset <- sample(1:dim(alldata)[1],subset.size)
  mva.pairs(alldata[subset,],labels=labels[chips],span=span,family.loess=family.loess,digits=digits,line.col=line.col,main=main,...)
} 

mva.pairs <- function(x,labels=colnames(x),log.it=T,span=2/3,family.loess="symmetric",digits=3,line.col=2,main="MVA plot",...){
  if(log.it) x <-log2(x)
  J <- dim(x)[2]
  frame()
  old.par <- par(no.readonly = TRUE)
   on.exit(par(old.par))
par(mfrow=c(J,J),mgp=c(0,.2,0),mar=c(1,1,1,1),oma=c(1,1.4,2,1))
  for(j in 1:(J-1)){
    par(mfg=c(j,j));plot(1,1,type="n",xaxt="n",yaxt="n",xlab="",ylab="");text(1,1,labels[j],cex=2)
    for(k in (j+1):J){
      par(mfg=c(j,k))
      yy <- x[,j]-x[,k]
      xx <-(x[,j]+x[,k])/2
      xx <- xx
      yy <- yy
      aux <- loess(yy~xx,degree=1,span=span,family=family.loess)$fitted
      plot(xx,yy,pch=".",xlab="",ylab="",tck=0,...)
      o <- order(xx)
      lines(xx[o],aux[o],col=line.col)
      par(mfg=c(k,j))
      sigma <- quantile(yy,.75)-quantile(yy,.25)
      txt <- format(c(sigma,0.123456789),digits=digits)
      plot(c(0,1),c(0,1),type="n",ylab="",xlab="",xaxt="n",yaxt="n")
      text(0.5,0.5,txt,cex=2)
    }
  }
  par(mfg=c(J,J));plot(1,1,type="n",xaxt="n",yaxt="n",xlab="",ylab="");
  text(1,1,labels[J],cex=2)
  mtext("A",1,outer=T,cex=1.5)
  mtext("M",2,outer=T,cex=1.5,las=1)
  mtext(main,3,outer=T,cex=1.5)
  invisible()
}

normalize <- function(x,method=c("quantile","loess","contrasts"),span=2/3,choose.subset=T,subset.size=5000,verbose=T,maxit=1){
  if(!is.element(method,c("quantile","robust.quantile","loess","contrasts")))
    stop(paste(method,"not a method!\n"))
  alldata <- rbind(x$pm,x$mm)
  index1 <- 1:x$nprob
  index2 <- (x$nprob+1):(2*x$nprob)

  if(method=="quantile") aux <- normalize.quantiles(alldata)
  if(method=="robust.quantile") aux <- normalize.quantiles.robust(alldata)
  ###robust.quantile is experimental, therefore no option is provided
  if(method=="loess"){
    if(choose.subset)
      subset1 <- maffy.subset(alldata,verbose=verbose,subset.size=subset.size)$subset
    else
      subset1 <- sample(1:dim(alldata)[1],subset.size)
    aux <-  loess.normalize(alldata,subset=subset1,maxit=maxit,verbose=verbose,span=span)
  }
  if(method=="contrasts"){
    if(choose.subset)
      subset1 <- maffy.subset(alldata,verbose=verbose,subset.size=subset.size)$subset
    else
      subset1 <- sample(1:dim(alldata)[1],subset.size)
  aux <-   maffy.normalise(alldata,subset=subset1,verbose=verbose,span=span)
  }
  x$pm <- aux[index1,]
  x$mm <- aux[index2,]
  x
}

loess.normalize <- function(mat,subset=sample(1:(dim(mat)[2]),5000),
                      epsilon=10^-2,maxit=1,log.it=T,verbose=T,span=2/3,
                      family.loess="symmetric"){
  J <- dim(mat)[2]
  II <- dim(mat)[1]
  newData <- mat
  if(log.it){
    mat <- log2(mat)
    newData <- log2(newData)
  }
  change <- epsilon +1
  fs <- matrix(0,II,J)##contains what we substract
  iter <- 0
  w <- c(0,rep(1,length(subset)),0) ##this way we give 0 weight to the
  ##extremes added so that we can interpolate
  while(iter < maxit){
    iter <- iter+1
    means <- matrix(0,II,J) ##contains temp of what we substract
    for(j in 1:(J-1)){
      for(k in (j+1):J){
        y <- newData[,j]-newData[,k]
        x <-(newData[,j]+newData[,k])/2
        index <- c(order(x)[1],subset,order(-x)[1])
        ##put endpoints in so we can interpolate
        xx <- x[index]
        yy <- y[index]
        aux <-loess(yy~xx,span=span,degree=1,weights=w,family=family.loess)
        aux <- predict.loess(aux,data.frame(xx=x))/J
        means[,j] <- means[,j] + aux 
        means[,k] <- means[,k] - aux
        if(verbose) cat("Done with",j,"vs",k," in iteration ",iter,"\n")
      }
    }
    fs <- fs+means
    newData <- mat-fs
    change <- max(apply((means[subset,])^2,2,mean))
    if(verbose) cat(iter,change,"\n")
    oldfs <- fs
  }
  if(change>epsilon & maxit>1) warning(paste("No convergence after",maxit,"iterations.\n"))
  if(log.it) return(2^newData)
  else return(newData)
}


ratio.hist <- function(x,j,prob=c(0,.25,.75,.95,1),
                       nclass=rep(75,length(prob)-1)){
  lpm <- log2(x$pm[,j])
  lmm <- log2(x$mm[,j])
  auxx <- quantile(lpm+lmm,prob)
  for(i in 1:(length(auxx)-1)){
    hist((lpm-lmm)[(lpm+lmm)>=auxx[i] & (lpm+lmm)<=auxx[i+1]],nclass=nclass[i],xlab="",xlim=range(lpm-lmm),main=paste(round(prob[i]*100),"% - ",round(prob[i+1]*100),"%",sep=""))
  }
  NULL
}

probe.profile <- function(x,probe.set.id,chips=1:x$nchips,by.chip=T,
                          pm.col="green",mm.col="red",global.ylim=T,...){
  index <- (1:x$nprob)[x$id==probe.set.id]
  probes <- x$numbers[index]
  o <- order(probes)
  index <- index[o]
  probes <- probes[o]
  GYLIM <- log2(range(rbind(x$pm[index,chips],x$mm[index,chips])))
  if(by.chip){
    for(i in chips){
      p <- log2(x$pm[index,i])
      m <- log2(x$mm[index,i])
      if(global.ylim) YLIM <- GYLIM else YLIM <- range(c(p,m))
      plot(probes,p,type="n",ylim=YLIM,
           xlab="probe number",ylab="log intensity",
           main=x$chip.names[i],...)
      points(probes,p,pch="p",col=pm.col,type="b")
      points(probes,m,pch="m",col=mm.col,type="b")
    }
  }
  else{
    for(i in 1:length(index)){
      p <- log2(x$pm[index[i],chips])
      m <- log2(x$mm[index[i],chips])
      if(global.ylim) YLIM <- GYLIM else YLIM <- range(c(p,m))
      plot(chips,p,type="n",ylim=YLIM,
           xlab="chip",ylab="log intensity",
           main=paste("probe",as.character(probes[i])),xaxt="n")
      axis(1,at=chips,labels=x$chip.names)
      points(chips,p,pch="p",col=pm.col,type="b")
      points(chips,m,pch="m",col=mm.col,type="b")
    }
  }
}
   
hlog <- function(x,constant=1){ #constant is where the change occurs
  if(constant<=0){
    warning("constant less than or equal to 0. Returning log(x)\n")
    return(log(x))
  }
  else{
    if(constant==Inf)
      return(x)
    else{
      aux <- (abs(x)<constant)*(x) +
        (abs(x)>=constant)*(sign(x)*(constant*log(abs(x/constant))+constant))
      aux[x==0] <- 0
      return(aux)
    }
  }
}

express <- function(x,method=c("avglogpm","li.wong","avdiff"),
                    normalize=T,
                    normalize.method=c("quantile","loess","contrasts"),
                    span=2/3,choose.subset=T,subset.size=5000,
                    verbose=T,maxit=1){
  if(!is.element(method,c("avglogpm","li.wong","avdiff")))
    stop(paste(method,"not a method!\n"))
  if(verbose) cat("Performing",normalize.method[1],"normalization.\n")
  if(normalize) x <- normalize(x,method=normalize.method,span=span,choose.subset=choose.subset,subset.size=subset.size,verbose=verbose,maxit=maxit)
  if(verbose) cat("Computing",method[1],"expression\n")
  if(method=="avglogpm"){
    bg <- apply(log2(x$mm),2,function(x){
      aux <- density(x,kernel="epanechnikov",n=2^16)
      2^aux$x[order(-aux$y)[1]]})
    aux <- sweep(x$pm,2,bg)
    aux[aux<=0] <- min(aux[aux>0])/2
    aux <- log2(aux)
    aux <- apply(aux,2,function(y) tapply(y,x$id,mean))
    dimnames(aux) <- list(x$names,x$chip.names)
    return(2^aux)
  }
  if(method=="li.wong"){
    aux <- li.wong(x)$theta
    dimnames(aux) <- list(x$names,x$chip.names)
    return(aux)
  }
  if(method=="avdiff") return(avdiff(x))
}


